@@ -37,8 +37,11 @@
 #include <libgnome/gnome-desktop-item.h>
 #include <libgnomevfs/gnome-vfs-utils.h>
 
+#include <errno.h>
+#include <fcntl.h>
 #include <string.h> /* for strcmp */
 #include <unistd.h> /* for chdir */
+#include <sys/stat.h>
 
 #define SSH_DEFAULT_PORT 22
 
@@ -58,17 +61,21 @@ static TerminalFileInfo
 get_terminal_file_info (NautilusFileInfo *file_info)
 {
 	TerminalFileInfo  ret;
-	gchar            *uri_scheme;
+	char             *uri_scheme, *p;
 
 	g_assert (file_info);
 
-	uri_scheme = nautilus_file_info_get_uri_scheme (file_info);
+	uri_scheme = nautilus_file_info_get_activation_uri (file_info);
+	if (p = strchr (uri_scheme, ':')) {
+		*p = 0;
+	}
 
 	if (strcmp (uri_scheme, "file") == 0) {
 		ret = FILE_INFO_LOCAL;
 	} else if (strcmp (uri_scheme, "x-nautilus-desktop") == 0) {
 		ret = FILE_INFO_DESKTOP;
-	} else if (strcmp (uri_scheme, "sftp") == 0) {
+	} else if (strcmp (uri_scheme, "sftp") == 0 ||
+		   strcmp (uri_scheme, "ssh") == 0) {
 		ret = FILE_INFO_SFTP;
 	} else {
 		ret = FILE_INFO_OTHER;
@@ -117,6 +124,24 @@ lookup_in_data_dirs (const char *basename)
 	return NULL;
 }
 
+static inline gboolean
+desktop_opens_home_dir (GConfClient *client)
+{
+	return gconf_client_get_bool (client, "/apps/nautilus-open-terminal/desktop_opens_home_dir", NULL);
+}
+
+static inline gboolean
+desktop_is_home_dir (GConfClient *client)
+{
+	return gconf_client_get_bool (client, "/apps/nautilus/preferences/desktop_is_home_dir", NULL);
+}
+
+#ifdef HAVE_GLIB_DESKTOP_DIR_API
+  #define get_desktop_dir() g_strdup (g_get_user_special_dir (G_USER_DIRECTORY_DESKTOP))
+#else
+  #define get_desktop_dir() g_build_filename (g_get_home_dir (), "Desktop", NULL)
+#endif
+
 static void
 append_sftp_info (char **terminal_exec,
 		  NautilusFileInfo *file_info)
@@ -129,8 +154,13 @@ append_sftp_info (char **terminal_exec,
 	g_assert (terminal_exec != NULL);
 	g_assert (file_info != NULL);
 
-	uri = nautilus_file_info_get_uri (file_info);
+	uri = nautilus_file_info_get_activation_uri (file_info);
+	g_assert (uri != NULL);
+	g_assert (strncmp (uri, "sftp", strlen ("sftp")) == 0 ||
+		  strncmp (uri, "ssh", strlen ("ssh")) == 0);
+
 	vfs_uri = gnome_vfs_uri_new (uri);
+	g_assert (vfs_uri != NULL);
 
 	host_name = gnome_vfs_uri_get_host_name (vfs_uri);
 	host_port = gnome_vfs_uri_get_host_port (vfs_uri);
@@ -149,7 +179,7 @@ append_sftp_info (char **terminal_exec,
 		user_host = g_strdup (host_name);
 	}
 
-	cmd = g_strdup_printf ("ssh %s -p %d -t \"cd %s && sh -l\"", user_host, host_port, unescaped_path);
+	cmd = g_strdup_printf ("ssh %s -p %d -t \"cd \'%s\' && $SHELL -l\"", user_host, host_port, unescaped_path);
 	quoted_cmd = g_shell_quote (cmd);
 	g_free (cmd);
 
@@ -185,13 +215,14 @@ open_terminal_callback (NautilusMenuItem *item,
 						 "exec",
 						 NULL);
 
-	if (terminal_exec == NULL) {
+	if (terminal_exec == NULL || strlen (terminal_exec) == 0) {
+		g_free (terminal_exec);
 		terminal_exec = g_strdup ("gnome-terminal");
 	}
 
 	switch (get_terminal_file_info (file_info)) {
 		case FILE_INFO_LOCAL:
-			uri = nautilus_file_info_get_uri (file_info);
+			uri = nautilus_file_info_get_activation_uri (file_info);
 			if (uri != NULL) {
 				working_directory = g_filename_from_uri (uri, NULL, NULL);
 			} else {
@@ -201,7 +232,11 @@ open_terminal_callback (NautilusMenuItem *item,
 			break;
 
 		case FILE_INFO_DESKTOP:
-			working_directory = g_strdup (g_get_home_dir ());
+			if (desktop_is_home_dir (client) || desktop_opens_home_dir (client)) {
+				working_directory = g_strdup (g_get_home_dir ());
+			} else {
+				working_directory = get_desktop_dir ();
+			}
 			break;
 
 		case FILE_INFO_SFTP:
@@ -232,17 +267,61 @@ open_terminal_callback (NautilusMenuItem *item,
 	}
 
 	if (dfile != NULL) {
-		if (working_directory != NULL) {
-			chdir (working_directory);
+		int orig_cwd = -1;
+
+		do {
+			orig_cwd = open (".", O_RDONLY);
+		} while (orig_cwd == -1 && errno == EINTR);
+
+		if (orig_cwd == -1) {
+			g_message ("NautilusOpenTerminal: Failed to open current Nautilus working directory.");
+		} else if (working_directory != NULL) {
+
+			if (chdir (working_directory) == -1) {
+				int ret;
+
+				g_message ("NautilusOpenTerminal: Failed to change Nautilus working directory to \"%s\".",
+					   working_directory);
+
+				do {
+					ret = close (orig_cwd);
+				} while (ret == -1 && errno == EINTR);
+
+				if (ret == -1) {
+					g_message ("NautilusOpenTerminal: Failed to close() current Nautilus working directory.");
+				}
+
+				orig_cwd = -1;
+			}
 		}
 
 		ditem = gnome_desktop_item_new_from_file (dfile, 0, NULL);
 
 		gnome_desktop_item_set_string (ditem, "Exec", terminal_exec);
-		gnome_desktop_item_set_launch_time (ditem, gtk_get_current_event_time ());
+		if (gtk_get_current_event_time () > 0) {
+			gnome_desktop_item_set_launch_time (ditem, gtk_get_current_event_time ());
+		}
 		gnome_desktop_item_launch (ditem, NULL, GNOME_DESKTOP_ITEM_LAUNCH_USE_CURRENT_DIR, NULL);
 		gnome_desktop_item_unref (ditem);
 		g_free (dfile);
+
+		if (orig_cwd != -1) {
+			int ret;
+
+			ret = fchdir (orig_cwd);
+			if (ret == -1) {
+				g_message ("NautilusOpenTerminal: Failed to change back Nautilus working directory to original location after changing it to \"%s\".",
+					   working_directory);
+			}
+
+			do {
+				ret = close (orig_cwd);
+			} while (ret == -1 && errno == EINTR);
+
+			if (ret == -1) {
+				g_message ("NautilusOpenTerminal: Failed to close Nautilus working directory.");
+			}
+		}
 	} else {	
 		g_spawn_async (working_directory,
 			       argv,
@@ -283,8 +362,13 @@ open_terminal_menu_item_new (TerminalFileInfo  terminal_file_info,
 			break;
 
 		case FILE_INFO_DESKTOP:
-			name = _("Open _Terminal");
-			tooltip = _("Open a terminal");
+			if (desktop_opens_home_dir (gconf_client_get_default ())) {
+				name = _("Open _Terminal");
+				tooltip = _("Open a terminal");
+			} else {
+				name = _("Open In _Terminal");
+				tooltip = _("Open the currently open folder in a terminal");
+			}
 			break;
 
 		case FILE_INFO_OTHER:
@@ -315,6 +399,9 @@ nautilus_open_terminal_get_background_items (NautilusMenuProvider *provider,
 		case FILE_INFO_DESKTOP:
 		case FILE_INFO_SFTP:
 			item = open_terminal_menu_item_new (terminal_file_info, gtk_widget_get_screen (window), FALSE);
+			g_object_set_data_full (G_OBJECT (item), "file-info",
+						g_object_ref (file_info),
+						(GDestroyNotify) g_object_unref);
 			g_signal_connect (item, "activate",
 					  G_CALLBACK (open_terminal_callback),
 					  file_info);
@@ -338,15 +425,20 @@ nautilus_open_terminal_get_file_items (NautilusMenuProvider *provider,
 	TerminalFileInfo  terminal_file_info;
 
 	if (g_list_length (files) != 1 ||
-	    !nautilus_file_info_is_directory (files->data)) {
+	    (!nautilus_file_info_is_directory (files->data) &&
+	     nautilus_file_info_get_file_type (files->data) != G_FILE_TYPE_SHORTCUT &&
+	     nautilus_file_info_get_file_type (files->data) != G_FILE_TYPE_MOUNTABLE)) {
 		return NULL;
 	}
 
 	terminal_file_info = get_terminal_file_info (files->data);
-	switch (get_terminal_file_info (files->data)) {
+	switch (terminal_file_info) {
 		case FILE_INFO_LOCAL:
 		case FILE_INFO_SFTP:
 			item = open_terminal_menu_item_new (terminal_file_info, gtk_widget_get_screen (window), TRUE);
+			g_object_set_data_full (G_OBJECT (item), "file-info",
+						g_object_ref (files->data),
+						(GDestroyNotify) g_object_unref);
 			g_signal_connect (item, "activate",
 					  G_CALLBACK (open_terminal_callback),
 					  files->data);
